'use strict';
var generators = require('yeoman-generator');
var chalk = require('chalk');
var yosay = require('yosay');
var _s = require('underscore.string');
var _uuid = require('node-uuid');
var camelCase = require('camelcase');
var removeDiacritics = require('diacritics').remove;
var hasBusinessKey = false;
var addedFields = [];
var className;
var classLabel;

const RESERVED_WORDS_JAVA = ["ABSTRACT", "CONTINUE", "FOR", "NEW", "SWITCH", "ASSERT", "DEFAULT", "GOTO", "PACKAGE", "SYNCHRONIZED", "BOOLEAN", "DO", "IF", "PRIVATE", "THIS", "BREAK", "DOUBLE", "IMPLEMENTS", "PROTECTED", "THROW", "BYTE", "ELSE", "IMPORT", "PUBLIC", "THROWS", "CASE", "ENUM", "INSTANCEOF", "RETURN", "TRANSIENT", "CATCH", "EXTENDS", "INT", "SHORT", "TRY", "CHAR", "FINAL", "INTERFACE", "STATIC", "VOID", "CLASS", "FINALLY", "LONG", "STRICTFP", "VOLATILE", "CONST", "FLOAT", "NATIVE", "SUPER", "WHILE"];

const RESERVED_WORDS_POSTGRES = ["ALL", "ANALYSE", "ANALYZE", "AND", "ANY", "ARRAY", "AS", "ASC", "ASYMMETRIC", "AUTHORIZATION", "BINARY", "BOTH", "CASE", "CAST", "CHECK", "COLLATE", "COLLATION", "COLUMN", "CONCURRENTLY", "CONSTRAINT", "CREATE", "CROSS", "CURRENT_CATALOG", "CURRENT_DATE", "CURRENT_ROLE", "CURRENT_SCHEMA", "CURRENT_TIME", "CURRENT_TIMESTAMP", "CURRENT_USER", "DEFAULT", "DEFERRABLE", "DESC", "DISTINCT", "DO", "ELSE", "END", "EXCEPT", "FALSE", "FETCH", "FOR", "FOREIGN", "FROM", "FULL", "GRANT", "GROUP", "HAVING", "ILIKE", "IN", "INITIALLY", "INNER", "INTERSECT", "INTO", "IS", "ISNULL", "JOIN", "LATERAL", "LEADING", "LEFT", "LIKE", "LIMIT", "LOCALTIME", "LOCALTIMESTAMP", "NATURAL", "NOT", "NOTNULL", "NULL", "OFFSET", "ON", "ONLY", "OR", "ORDER", "OUTER", "OVERLAPS", "PLACING", "PRIMARY", "REFERENCES", "RETURNING", "RIGHT", "SELECT", "SESSION_USER", "SIMILAR", "SOME", "SYMMETRIC", "TABLE", "THEN", "TO", "TRAILING", "TRUE", "UNION", "UNIQUE", "USER", "USING", "VARIADIC", "VERBOSE", "WHEN", "WHERE", "WINDOW", "WITH"];

const SOURCE_SRC_DIRECTORY = 'webpublico/src/main/java/br/com/webpublico/';

const SOURCE_TEST_DIRECTORY = 'webpublico/src/test/java/br/com/webpublico/';

const SOURCE_WEBAPP_DIRECTORY = 'webpublico/src/main/webapp/';

const SOURCE_RESOURCE_DIRECTORY = 'webpublico/src/main/resources/';

const DESTINATION_SRC_DIRECTORY = 'src/main/java/br/com/webpublico/';

const DESTINATION_TEST_DIRECTORY = 'src/test/java/br/com/webpublico/';

const DESTINATION_RESOURCE_DIRECTORY = 'src/main/resources/br/com/webpublico/';

const DESTINATION_WEBAPP_DIRECTORY = 'src/main/webapp/';

module.exports = generators.Base.extend({
    constructor: function() {
        generators.Base.apply(this, arguments);
        this.argument('className', {type: String, required: false});
        if (this.className) {
            this.appname = formatClassName(this.className);
        }
        this.option('noui');
        this.generateUI = !this.options.noui == true;
    },
    _askForField: function (cb) {
        this.fieldId++;
        var fieldNamesUnderscored = [];
        var prompts = [
            {
                type: 'input',
                name: 'label',
                message: 'Etiqueta do Campo (vazio para sair)',
                validate: function(input) {
                    var formattedInput = formatLabel(input);
                    if (formattedInput == 'id' || formattedInput == 'versao') {
                        return 'Os campos \'id\' e \'versao\' são adicionados automaticamente';
                    } else if (RESERVED_WORDS_JAVA.indexOf(formattedInput.toUpperCase()) != -1) {
                        return 'O nome do campo não pode conter palavras reservadas do Java';
                    } else if (RESERVED_WORDS_POSTGRES.indexOf(formattedInput.toUpperCase()) != -1) {
                        return 'O nome do campo não pode conter palavras reservadas do PostgreSQL';
                    }
                    return true;
                }
            },
            {
                when: function (response) {
                    return response.label && response.label != '';
                },
                type: 'input',
                name: 'fieldName',
                default: function(answer) {
                    return formatLabel(answer.label);
                },
                validate: function (input) {
                    if (!input) {
                        return true;
                    }
                    if (!(/^([a-zA-Z0-9_]*)$/.test(input))) {
                        return 'O nome do campo não pode conter Caracteres Especiais';
                    } else if (input == '') {
                        return 'O nome do campo não pode ser vazio';
                    } else if (input == 'id' || input == 'versao') {
                        return 'Já existe um campo com esse nome';
                    } else if (RESERVED_WORDS_JAVA.indexOf(input.toUpperCase()) != -1) {
                        return 'O nome do campo não pode conter palavras reservadas do Java';
                    } else if (RESERVED_WORDS_POSTGRES.indexOf(input.toUpperCase()) != -1) {
                        return 'O nome do campo não pode conter palavras reservadas do PostgreSQL';
                    }
                    return true;
                },
                message: 'Nome do Campo',
                filter: function(answer) {
                    return formatLabel(answer);
                }
            },
            {
                when: function (response) {
                    return response.label && response.label != '';
                },
                type: 'list',
                name: 'fieldType',
                message: 'Qual o Tipo do Campo?',
                choices: [
                    {
                        value: {
                            type: 'String',
                            size: 70,
                            isString: true
                        },
                        name: 'String Média - 70 Caracteres'
                    },
                    {
                        value: {
                            type: 'Integer',
                            isNumber: true,
                            isInteger: true
                        },
                        name: 'Integer'
                    },
                    {
                        value: {
                            type: 'Date',
                            isDate: true
                        },
                        name: 'Date'
                    },
                    {
                        value: {
                            type: 'BigDecimal',
                            size: 2,
                            isNumber: true,
                            isDecimal: true
                        },
                        name: 'BigDecimal - 2 Casas Decimais'
                    },
                    {
                        value: {
                            type: 'Boolean'
                        },
                        name: 'Boolean'
                    },
                    {
                        value: {
                            type: 'String',
                            size: 20,
                            isString: true
                        },
                        name: 'String Curta - 20 Caracteres'
                    },
                    {
                        value: {
                            type: 'String',
                            size: 255,
                            isString: true
                        },
                        name: 'String Longa - 255 Caracteres'
                    },
                    {
                        value: {
                            type: 'String',
                            size: 3000,
                            isString: true
                        },
                        name: 'String Extra Longa - 3000 Caracteres'
                    },
                    {
                        value: {
                            type: 'String',
                            size: 0,
                            isString: true
                        },
                        name: 'String Infinita - Texto'
                    },
                    {
                        value: {
                            type: 'BigDecimal',
                            size: 4,
                            isNumber: true,
                            isDecimal: true
                        },
                        name: 'BigDecimal - 4 Casas Decimais'
                    },
                    {
                        value: {
                            type: 'Long',
                            isNumber: true,
                            isLong: true
                        },
                        name: 'Long'
                    },
                    {
                        value: {
                            type: 'Date',
                            isTime: true
                        },
                        name: 'Timestamp'
                    },
                    {
                        value: {
                            type: 'LocalDate'
                        },
                        name: 'LocalDate (Date Java 8)'
                    },
                    {
                        value: {
                            type: 'ZonedDateTime'
                        },
                        name: 'ZonedDateTime (DateTime Java 8)'
                    }
                ],
                default: 0
            },
            {
                when: function (response) {
                    return response.label && response.label != '';
                },
                type: 'checkbox',
                name: 'fieldAnnotations',
                message: 'Validações',
                choices: [
                    {name: 'Obrigatório', value: 'obrigatorio'},
                    {name: 'Tabelável', value: 'tabelavel'},
                    {name: 'Pesquisável', value: 'pesquisavel'}
                ],
                default: ['tabelavel', 'pesquisavel']
            },
            {
                when: function (response) {
                    return false;
                    /*return response.fieldAdd
                        && response.fieldType.isNumber;*/
                },
                type: 'checkbox',
                name: 'numberMinValue',
                message: 'Validações do Campo',
                choices: [
                    {
                        value: 'gtZero',
                        name: 'Maior que Zero'
                    },
                    {
                        value: 'geZero',
                        name: 'Maior ou Igual a Zero'
                    }
                ]
            }
        ];
        this.prompt(prompts, function (props) {
            //if (props.fieldAdd) {
            if (props.label && props.label != '') {
                var field = {
                    fieldId: this.fieldId,
                    fieldName: formatVariableName(props.fieldName),
                    requiredField: props.requiredField,
                    capitalizedFieldName: formatClassName(props.fieldName),
                    underscoredFieldName: formatUrlName(props.fieldName, "_"),
                    fieldType: props.fieldType,
                    liquibaseFieldType: getLiquibaseFieldType(props.fieldType.type),
                    label: props.label,
                    annotations: {
                        obrigatorio: props.fieldAnnotations.indexOf('obrigatorio') > -1,
                        pesquisavel: props.fieldAnnotations.indexOf('pesquisavel') > -1,
                        tabelavel: props.fieldAnnotations.indexOf('tabelavel') > -1
                    },
                    validations: {
                        required: props.requiredField,
                        number: {
                            minValue: props.numberMinValue
                        }
                    }
                };
                if (field.annotations.obrigatorio == true) {
                    this.fieldsObrigatorios.push(field);
                }
                if (field.fieldType.type == 'Date') {
                    this.temCampoDate = true;
                }
                if (field.fieldType.type == 'LocalDate') {
                    this.temCampoLocalDate = true;
                }
                if (field.fieldType.type == 'ZonedDateTime') {
                    this.temCampoZonedDateTime = true;
                }
                fieldNamesUnderscored.push(_s.underscored(props.fieldName));
                this.fields.push(field);
                addedFields.push(field);
                this._askForField(cb);
            } else {
                cb();
            }
        }.bind(this));
    },
    initializing: {
        inicializandoVariaveis: function () {
            this.fieldId = 0;
            this.fields = [];
            this.fieldsObrigatorios = [];
            this.chaveNegocioFields = [];
            this.temCampoDate = false;
            this.temCampoLocalDate = false;
            this.temCampoZonedDateTime = false;
        }
    },
    writing: {
        config: function () {
            var srcDestination = DESTINATION_SRC_DIRECTORY + this.modulo + '/';
            var testDestination = DESTINATION_TEST_DIRECTORY + this.modulo + '/';
            var webAppDestination = DESTINATION_WEBAPP_DIRECTORY + this.modulo + '/';
            var recursoSistemaDestination = DESTINATION_RESOURCE_DIRECTORY + 'recursosistema/' + this.modulo + '/';
            this.template(SOURCE_SRC_DIRECTORY + 'entities/_Entity.java', srcDestination + 'entities/' + this.classe + '.java');
            this.template(SOURCE_TEST_DIRECTORY + 'entities/_EntityTest.java', testDestination + 'entities/' + this.classe + 'Test.java');
            if (this.temChaveNegocio) {
                this.template(SOURCE_SRC_DIRECTORY + 'chavesnegocio/_ChaveNegocio.java', srcDestination + 'chavesnegocio/ChaveNegocio' + this.classe + '.java');
            }
            this.template(SOURCE_SRC_DIRECTORY + 'repositories/jpa/_JPARepository.java', srcDestination + 'repositories/jpa/' + this.classe + 'JPARepository.java');
            this.template(SOURCE_TEST_DIRECTORY + 'repositories/jpa/_JPARepositoryTest.java', testDestination + 'repositories/jpa/' + this.classe + 'JPARepositoryTest.java');
            this.template(SOURCE_SRC_DIRECTORY + 'repositories/jdbc/_JDBCRepository.java', srcDestination + 'repositories/jdbc/' + this.classe + 'JDBCRepository.java');
            this.template(SOURCE_SRC_DIRECTORY + 'repositories/jdbc/statements/_Statements.java', srcDestination + 'repositories/jdbc/statements/Statements' + this.classe + '.java');
            this.template(SOURCE_SRC_DIRECTORY + 'services/_SpringService.java', srcDestination + 'services/' + this.classe + 'Service.java');
            this.template(SOURCE_TEST_DIRECTORY + 'services/_SpringServiceTest.java', testDestination + 'services/' + this.classe + 'ServiceTest.java');
            this.template(SOURCE_TEST_DIRECTORY + 'services/_SpringServiceIT.java', testDestination + 'services/' + this.classe + 'ServiceIT.java');
            if (this.generateUI) {
                this.template(SOURCE_SRC_DIRECTORY + 'controllers/_JSFController.java', srcDestination + 'controllers/' + this.classe + 'Controller.java');
                this.template(SOURCE_WEBAPP_DIRECTORY + '_lista.xhtml', webAppDestination + this.url + '/lista.xhtml');
                this.template(SOURCE_WEBAPP_DIRECTORY + '_edita.xhtml', webAppDestination + this.url + '/edita.xhtml');
                this.template(SOURCE_WEBAPP_DIRECTORY + '_visualizar.xhtml', webAppDestination + this.url + '/visualizar.xhtml');
                this.template(SOURCE_RESOURCE_DIRECTORY + '/recursosistema/_lista.json', recursoSistemaDestination + this.url + '/lista.json');
                this.template(SOURCE_RESOURCE_DIRECTORY + '/recursosistema/_edita.json', recursoSistemaDestination + this.url + '/edita.json');
                this.template(SOURCE_RESOURCE_DIRECTORY + '/recursosistema/_visualizar.json', recursoSistemaDestination + this.url + '/visualizar.json');
            }
        }
    },
    prompting: {
        perguntaTabela: function () {
            var cb = this.async();
            var prompts = [{
                type: 'input',
                name: 'classLabel',
                message: 'Etiqueta da Classe',
                default: this.appname,
                filter: function(answer) {
                    classLabel = formatClassName(formatLabel(answer));
                    return answer;
                }
            },
                {
                type: 'input',
                name: 'className',
                message: 'Nome da Entidade',
                validate: function (input) {
                    if (!(/^([a-zA-Z0-9_]*)$/.test(input))) {
                        return 'O Nome da Entidade não pode conter Caracteres Especiais';
                    } else if (input == '') {
                        return 'O Nome da Entidade não pode ser vazio';
                    }
                    return true;
                },
                default: function() {
                    return classLabel;
                },
                filter: function (answer) {
                    className = formatClassName(answer);
                    return className;
                }
            },
                {
                    type: 'input',
                    name: 'classPlural',
                    message: 'Nome da Classe no Plural',
                    default: function() {
                        return className + "s";
                    }
                },
                {
                    type: 'list',
                    name: 'classGender',
                    message: 'Gênero da Entidade',
                    choices: [
                        {
                            value: 'masculino',
                            name: 'Masculino'
                        },
                        {
                            value: 'feminino',
                            name: 'Feminino'
                        }
                    ],
                    default: 'masculino'
                },
                {
                    type: 'list',
                    name: 'moduleName',
                    message: 'Módulo',
                    choices: [
                        {
                            value: 'comum',
                            name: 'Comum - Caso de Uso pertence a mais de um módulo'
                        },
                        {
                            value: 'administrativo',
                            name: 'Administrativo'
                        },
                        {
                            value: 'contabil',
                            name: 'Contábil'
                        },
                        {
                            value: 'rh',
                            name: 'RH - Recursos Humanos'
                        },
                        {
                            value: 'tributario',
                            name: 'Tributário'
                        }
                    ]
                },
                {
                    type: 'confirm',
                    name: 'businessKey',
                    message: 'Possui Chave de Negócio?'
                }];
            this.prompt(prompts, function (answers) {
                this.uuid_lista = _uuid.v4();
                this.uuid_edita = _uuid.v4();
                this.uuid_visualizar = _uuid.v4();
                this.etiqueta = answers.classLabel;
                this.classe = formatClassName(answers.className);
                this.classePlural = formatClassName(answers.classPlural);
                this.genero = answers.classGender;
                if (answers.classGender == 'feminino') {
                    this.finalGenero = 'a';
                } else {
                    this.finalGenero = 'o';
                }
                this.modulo = answers.moduleName;
                this.pacote = formatPackageName(this.classe);
                this.temChaveNegocio = answers.businessKey;
                this.variavel = formatVariableName(this.classe);
                this.variavelPlural = formatVariableName(this.classePlural);
                this.url = formatUrlName(this.classe, "-");
                this.tabela = formatTableName(this.classe);
                this.log(answers);
                this.getRandomValue = getRandomValue;
                hasBusinessKey = answers.businessKey;
                cb()
            }.bind(this));
        },
        perguntaCampos: function () {
            var cb = this.async();
            this._askForField(cb);
        },
        perguntaChaveNegocio: function () {
            var cb = this.async();
            var prompts = [{
                when: function () {
                    return hasBusinessKey && addedFields.length > 0;
                },
                type: 'checkbox',
                name: 'businessKeyFields',
                message: 'Selecione os Campos da Chave de Negócio',
                default: [],
                choices: function () {
                    var retorno = [];
                    for (var i = 0; i < addedFields.length; i++) {
                        var field = addedFields[i];
                        retorno.push({'name': field.label, 'value': field});
                    }
                    return retorno;
                },
                validate: function(answer) {
                    if (!answer || answer.length <= 0) {
                        return "Ao Menos um Campo deve ser Selecionado";
                    }
                    return true;
                }
            }];
            this.prompt(prompts, function (answers) {
                this.chaveNegocioFields = answers.businessKeyFields;
                cb()
            }.bind(this));
        }
    }
});

function formatClassName(name) {
    if (name) {
        return name.charAt(0).toUpperCase() + name.slice(1);
    }
}

function formatVariableName(name) {
    if (name) {
        return name.charAt(0).toLowerCase() + name.slice(1);
    }
}

function formatPackageName(name) {
    return name.toLowerCase();
}

function formatUrlName(name, replacement) {
    return name.charAt(0).toLowerCase() + name.slice(1).replace(/([A-Z])/g, replacement + "$1").toLowerCase();
}

function formatTableName(name) {
    return name.charAt(0).toLowerCase() + name.slice(1).replace(/([A-Z])/g, "_" + "$1").toLowerCase();
}

function getLiquibaseFieldType(fieldType) {
    switch (fieldType) {
        case 'String':
            return 'DESCRIPTION';
        case 'BigDecimal':
            return 'MONETARY';
        case 'Date':
            return 'DATE';
        case 'Timestamp':
            return 'TIMESTAMP';
        case 'Double':
            return 'DECIMAL';
        case 'Float':
            return 'PERCENTAGE';
    }
    return fieldType.toUpperCase();
}

function formatLabel(name) {
    return removeDiacritics(camelCase(name)).replace(/\W/g, '');
}

function getRandomValue(field) {
    if (!field || !field.fieldType) {
        return 'campo ou tipo do campo nulo';
    }
    if (field.fieldType.isString) {
        return '"' + getRandomString(field.fieldType.size) + '"';
    }
    if (field.fieldType.isInteger) {
        return getRandomInt();
    }
    if (field.fieldType.isLong) {
        return getRandomInt() + "L";
    }
    if (field.fieldType.isDecimal) {
        return 'new BigDecimal("' + getRandomInt() + '.' + getRandomIntAte(field.fieldType.size == 2 ? '99' : '9999') + '")';
    }
}

function getRandomString(tamanho) {
    var retorno = "";
    var valores = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
    for(var i = 0; i < tamanho; i++) {
        retorno += valores.charAt(Math.floor(Math.random() * valores.length));
    }
    return retorno;
}

function getRandomIntAte(max) {
    return Math.floor(Math.random() * (max - 1)) + 1;
}

function getRandomInt() {
    return getRandomIntAte(32767);
}


